#!/usr/bin/env python3
# Copyright (c) 2016 Iotic Labs Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://github.com/Iotic-Labs/py-IoticAgent/blob/master/LICENSE
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import unicode_literals
from datetime import datetime, timedelta
from uuid import UUID
from warnings import warn

from . import Const
from .compat import (PY3, string_types, int_types, arg_checker, ensure_ascii, ensure_unicode, number_types,
                     raise_from, Sequence, Mapping, re_compile)


VALIDATION_LID_LEN = 64
VALIDATION_MIME_LEN = 64
VALIDATION_TIME_FMT = '%Y-%m-%dT%H:%M:%S.%fZ'
VALIDATION_META_FMT = {'n3', 'xml', 'turtle'}
VALIDATION_META_LANGUAGE = 2
VALIDATION_META_LABEL = 64
VALIDATION_META_COMMENT = 256
VALIDATION_META_TAG_MIN = 3
VALIDATION_META_TAG_MAX = 64
VALIDATION_META_URIREF_MAX = 128
VALIDATION_META_SEARCH_TEXT = 128
# In addition to plain strings, the following primitive types (with implied XSD equivalents) are supported
VALIDATION_META_PROP_AUTOTYPES = (int, float, bool)
VALIDATION_MAX_ENCODED_LENGTH = int(round(1024 * 64 * 0.98))  # (64k is current hard AMQP message limit)
VALIDATION_FOC_TYPES = frozenset((Const.R_FEED, Const.R_CONTROL))
# http://www.w3.org/TR/xmlschema-2/#built-in-datatypes
VALIDATION_META_VALUE_TYPES = frozenset((
    "anyURI",
    "base64Binary",
    "boolean",
    "byte",
    "date",
    "dateTime",
    "decimal",
    "double",
    "float",
    "int",
    "integer",
    "long",
    "negativeInteger",
    "nonNegativeInteger",
    "nonPositiveInteger",
    "positiveInteger",
    "short",
    "string",
    "time",
    "unsignedByte",
    "unsignedInt",
    "unsignedLong",
    "unsignedShort",
    # virtual (non-xsd) type for specifying IRI (URIRef) values
    "IRI"
))

_PATTERN_ASCII = re_compile(r'(?%s)^\S+$' % 'a' if PY3 else '')
_PATTERN_LEAD_TRAIL_WHITESPACE = re_compile(r'(?u)^\s.*|.*\s$')
_PATTERN_WHITESPACE = re_compile(r'(?u)^.*\s.*$')
_PATTERN_LANGUAGE = re_compile(r'(?%si)^[a-z]{2}$' % 'a' if PY3 else '')
_PATTERN_TAG = re_compile(r'(?u)^[\w.-]{%d,%d}$' % (VALIDATION_META_TAG_MIN, VALIDATION_META_TAG_MAX))
# For e.g. splitting search text into individual words. Do not pick words surrounded by non-whitespace characters. Allow
# both words and tags. Remote end will validate terms in more detail.
_PATTERN_SEARCH_TERMS = re_compile(r'(?u)(?<!\S)[\w-]+(?!\S)')
# for validating fqdn/ip and path of a url (cannot contain whitespace, minimum length)
_PATTERN_NON_WHITESPACE_3PLUS = re_compile(r'(?u)^\S{3}\S*$')

# Inverse of disallowed characters in e.g. RDF N3 notation
_VALID_URIREF_CHAR = r'[^<>"\'\s{}|\\^`]'

_PATTERN_URIREF_ABSOLUTE = re_compile(r"""(?ux)
    ^

    # E.g. http (2+ chars)
    (?P<scheme>{urichar}{urichar}+?)://

    # Domain or IP (3+ chars)
    (?P<netloc>{urichar}{{2}}{urichar}+?)

    # Path (2+ chars)
    /(?P<path>{urichar}{urichar}+?)

    # Last path item or fragment (2+ chars, relative predicate)
    ([#/](?P<relpred>{urichar}{urichar}+?))

    $
""".format(urichar=_VALID_URIREF_CHAR))

_LOCATION_SEARCH_ARGS = frozenset(('lat', 'long', 'radius'))


class Validation(object):  # pylint: disable=too-many-public-methods

    @staticmethod
    def check_convert_string(obj, name=None,
                             no_leading_trailing_whitespace=True,
                             no_whitespace=False,
                             no_newline=True,
                             as_tag=False,
                             min_len=1,
                             max_len=0):
        """Ensures the provided object can be interpreted as a unicode string, optionally with
           additional restrictions imposed. By default this means a non-zero length string
           which does not begin or end in whitespace."""
        if not name:
            name = 'Argument'
        obj = ensure_unicode(obj, name=name)
        if no_whitespace:
            if _PATTERN_WHITESPACE.match(obj):
                raise ValueError('%s cannot contain whitespace' % name)
        elif no_leading_trailing_whitespace and _PATTERN_LEAD_TRAIL_WHITESPACE.match(obj):
            raise ValueError('%s contains leading/trailing whitespace' % name)
        if (min_len and len(obj) < min_len) or (max_len and len(obj) > max_len):
            raise ValueError('%s too short/long (%d/%d)' % (name, min_len, max_len))
        if as_tag:
            if not _PATTERN_TAG.match(obj):
                raise ValueError('%s can only contain alphanumeric (unicode) characters, numbers and the underscore'
                                 % name)
        # whole words cannot contain newline so additional check not required
        elif no_newline and '\n' in obj:
            raise ValueError('%s cannot contain line breaks' % name)
        return obj

    @classmethod
    def lid_check_convert(cls, lid):
        return cls.check_convert_string(lid, 'lid', max_len=VALIDATION_LID_LEN)

    @classmethod
    def pid_check_convert(cls, lid):
        return cls.check_convert_string(lid, 'pid', max_len=VALIDATION_LID_LEN)

    @staticmethod
    def guid_check_convert(guid, allow_none=False):
        """Take a GUID in the form of hex string "32" or "8-4-4-4-12".

        Returns:
            Hex string "32" or raises ValueError: badly formed hexadecimal UUID string
        """
        if isinstance(guid, string_types):
            return ensure_unicode(UUID(guid).hex)
        elif guid is None and allow_none:
            return None
        else:
            raise ValueError('guid must be a string')

    @staticmethod
    def limit_offset_check(limit, offset):
        if not (isinstance(limit, int_types) and limit >= 0 and
                isinstance(offset, int_types) and offset >= 0):
            raise ValueError("limit/offset invalid")

    @staticmethod
    def metafmt_check_convert(fmt):
        fmt = ensure_unicode(fmt, name='fmt')
        if fmt not in VALIDATION_META_FMT:
            raise ValueError("metadata fmt must be one of of %s" % VALIDATION_META_FMT)
        return fmt

    @classmethod
    def tags_check_convert(cls, tags):
        """Accept one tag as string or multiple tags in list of strings.

        Returns:
            List (with tags in unicode form) or raises ValueError
        """
        # single string check comes first since string is also a Sequence
        if isinstance(tags, string_types):
            return [cls.__tag_check_convert(tags)]
        elif isinstance(tags, Sequence):
            if not tags:
                raise ValueError("Tag list is empty")
            return [cls.__tag_check_convert(tag) for tag in tags]
        else:
            raise ValueError("tags must be a single string or list of sequence of strings")

    @classmethod
    def __tag_check_convert(cls, tag):
        return cls.check_convert_string(tag, 'tags', no_whitespace=True, as_tag=True,
                                        min_len=VALIDATION_META_TAG_MIN, max_len=VALIDATION_META_TAG_MAX)

    @staticmethod
    def bool_check_convert(bname, barg):  # pylint: disable=unused-argument
        """Currently does no checking - passthrough for now in case people change their minds"""
        return bool(barg)

    @classmethod
    def lang_check_convert(cls, lang, allow_none=False, default=None):
        if lang is None:
            if allow_none:
                return None
            else:
                lang = default
        lang = ensure_unicode(lang, name='lang')
        if not _PATTERN_LANGUAGE.match(lang):
            raise ValueError('Language should only contain a-z characters')
        return lang

    @classmethod
    def mime_check_convert(cls, mime, allow_none=False):
        if mime is None and allow_none:
            return None
        else:
            mime = ensure_ascii(mime, name='mime')
            if 1 < len(mime) < VALIDATION_MIME_LEN:
                return mime
            else:
                raise ValueError('mime too long (%d)' % VALIDATION_MIME_LEN)

    __zeroOffsetOrNone = frozenset((timedelta(0), None))

    @classmethod
    def datetime_check_convert(cls, time, allow_none=False, require_utc=True, to_iso8601=True):
        if time is None and allow_none:
            return None
        if not isinstance(time, datetime):
            raise ValueError('datetime instance required')
        if require_utc:
            offset = time.utcoffset()
            if offset not in cls.__zeroOffsetOrNone:
                raise ValueError('datetime instance must be naive or have zero UTC offset')
        return ensure_unicode(time.strftime(VALIDATION_TIME_FMT)) if to_iso8601 else time

    @staticmethod
    def foc_check(foc):
        if foc not in VALIDATION_FOC_TYPES:
            raise ValueError("Resource type invalid expected one of %s" % VALIDATION_FOC_TYPES)

    @classmethod
    def label_check_convert(cls, label, allow_none=False):
        if label is None and allow_none:
            return None
        else:
            return cls.check_convert_string(label, 'label', max_len=VALIDATION_META_LABEL)

    @classmethod
    def comment_check_convert(cls, comment, allow_none=False):
        if comment is None and allow_none:
            return None
        else:
            return cls.check_convert_string(comment, 'comment', max_len=VALIDATION_META_COMMENT, no_newline=False)

    description_check_convert = comment_check_convert

    @classmethod
    def value_type_check_convert(cls, vtype):
        warn('Use object_type_check_convert instead', DeprecationWarning)
        return cls.object_type_check_convert(vtype)

    @staticmethod
    def object_type_check_convert(otype):
        otype = ensure_unicode(otype, name='otype')
        if otype not in VALIDATION_META_VALUE_TYPES:
            raise ValueError("object type not a valid xsd primitive (or derived) type name")
        return otype

    @classmethod
    def proplist_check_convert(cls, props, for_delete=False):
        """for_delete allows a completely empty list as well as no value for individual properties. This is for deletion
        of all properties or removal of all properties with a given predicate respectively.
        """
        if not isinstance(props, Sequence):
            raise ValueError('properties list (or single property) must be a sequence')
        if not props:
            if not for_delete:
                raise ValueError('properties list cannot be empty')
        # support single property being supplied
        elif isinstance(props[0], string_types):
            props = (props,)

        output = []

        for i, prop in enumerate(props):
            try:
                if not (isinstance(prop, Sequence) and 2 <= len(prop) <= 3):
                    raise ValueError('Must be sequence of length 2-3')
                predicate = cls.uriref_check_convert(prop[0], name='predicate')
                value = prop[1]
                # (predicate, object-value)
                if len(prop) == 2:
                    if value is None:
                        if not for_delete:
                            raise ValueError('Value cannot be None')
                        prop = (predicate, None)
                    elif isinstance(value, VALIDATION_META_PROP_AUTOTYPES):
                        prop = (predicate, value)
                    else:
                        prop = (predicate, ensure_unicode(value, 'value'))
                # (predicate, object-value, object-type)
                else:
                    prop = (predicate, ensure_unicode(value, 'value'), cls.object_type_check_convert(prop[2]))
            except ValueError as ex:
                raise_from(ValueError('Property #%d invalid: %s' % (i, ex)), ex)

            output.append(prop)

        return output

    @classmethod
    def value_unit_check_convert(cls, unit):
        warn('Use uriref_check_convert instead', DeprecationWarning)
        return cls.uriref_check_convert(unit)

    # TODO (question): Could enforce http(s) scheme as before, but is that necessary/wanted?
    @staticmethod
    def uriref_check_convert(ref, name='uriref'):
        """Performs minimal/sanity validation for RDF URI References."""
        ref = ensure_unicode(ref, name=name)
        if len(ref) > VALIDATION_META_URIREF_MAX:
            raise ValueError('%s too long (%d, max=%d)' % (name, len(ref), VALIDATION_META_URIREF_MAX))
        if not _PATTERN_URIREF_ABSOLUTE.match(ref):
            raise ValueError('%s malformed' % name)
        return ref

    @staticmethod
    def location_check(lat, lon):
        """For use by Core client wrappers"""
        if not (isinstance(lat, number_types) and -90 <= lat <= 90):
            raise ValueError("Latitude: '{latitude}' invalid".format(latitude=lat))

        if not (isinstance(lon, number_types) and -180 <= lon <= 180):
            raise ValueError("Longitude: '{longitude}' invalid".format(longitude=lon))

    @classmethod
    def search_location_check(cls, location):
        """Core.Client.request_search location parameter should be a dictionary that contains lat, lon and radius floats
        """
        if not (isinstance(location, Mapping) and set(location.keys()) == _LOCATION_SEARCH_ARGS):
            raise ValueError('Search location should be mapping with keys: %s' % _LOCATION_SEARCH_ARGS)

        cls.location_check(location['lat'], location['long'])
        radius = location['radius']
        if not (isinstance(radius, number_types) and 0 < radius <= 20038):  # half circumference
            raise ValueError("Radius: '{radius}' is invalid".format(radius=radius))

    @classmethod
    def search_check_convert(cls, text, lang, location, unit, default_lang):
        arg_count = 0
        payload = {}

        if text is not None:
            payload['text'] = cls.__search_text_check_convert(text)
            arg_count += 1

        payload['lang'] = cls.lang_check_convert(lang, default=default_lang)

        if location is not None:
            cls.search_location_check(location)
            payload.update(location)
            arg_count += 1
            if location['radius'] > 25 and text is None:
                raise ValueError('radius cannot exceed 25km when no search text supplied')

        if unit is not None:
            payload['unit'] = cls.uriref_check_convert(unit)
            arg_count += 1

        if not arg_count:
            raise ValueError('At least one of text, location and unit must be specified')

        return payload

    @classmethod
    def search_type_check_convert(cls, type_):
        return cls.__check_convert_str_enum(Const.SearchType, 'Search type', 'type_', type_)

    @classmethod
    def search_scope_check_convert(cls, type_):
        return cls.__check_convert_str_enum(Const.SearchScope, 'Search scope', 'scope', type_)

    @classmethod
    def describe_scope_check_convert(cls, type_):
        return cls.__check_convert_str_enum(Const.DescribeScope, 'Describe scope', 'scope', type_)

    @staticmethod
    def __check_convert_str_enum(enum_cls, enum_name, var_name, value):
        if isinstance(value, enum_cls):
            return value
        # Allow for old behaviour, specified as string
        try:
            return enum_cls(ensure_ascii(value, name=var_name))
        except ValueError as ex:
            raise_from(ValueError('%s must be one of: %s' % (enum_name, ', '.join(str(x) for x in enum_cls))), ex)

    @classmethod
    def __search_text_check_convert(cls, text):
        """Converts and keeps only words in text deemed to be valid"""
        text = cls.check_convert_string(text, name='text', no_leading_trailing_whitespace=False)
        if len(text) > VALIDATION_META_SEARCH_TEXT:
            raise ValueError("Search text can contain at most %d characters" % VALIDATION_META_SEARCH_TEXT)
        text = ' '.join(_PATTERN_SEARCH_TERMS.findall(text))
        if not text:
            raise ValueError('Search text must contain at least one non-whitespace term (word)')
        return text

    @staticmethod
    def callable_check(func, arg_count=1, arg_value=None, allow_none=False):
        """Check whether func is callable, with the given number of positional arguments.

        Returns:
            True if check succeeded, False otherwise."""
        if func is None:
            if not allow_none:
                raise ValueError('callable cannot be None')
        elif not arg_checker(func, *[arg_value for _ in range(arg_count)]):
            raise ValueError('callable %s invalid (for %d arguments)' % (func, arg_count))
